package XML::DT;

BEGIN{
 use XML::Parser;
 use Exporter ();
 @ISA=qw(Exporter);
 @EXPORT=qw(dt dtstring inctxt ctxt mkdtskel toxml);
 $VERSION = '0.08';
}

=head1 NAME

    XML::DT - a package for down translation of XML to strings

=head1 SYNOPSIS

    use XML::DT;

    %xml=( 'music'    => sub{"Music from: $c\n"},
           'lyrics'   => sub{"Lyrics from:$c\n (the value of attribute
                               IN is:$v{IN}\n)"},
           'title'    => sub{ uc($c) },
           '-default' => sub{"$q:$c"},
           '-outputenc' => 'ISO-8859-1');
    
    print dt($filename,%xml);

    print dtstring("<arq>
                    <title>Vejam Bem</title>
                    <music>Zeca Afonso</music>
                    </arq>",%xml);

    inctxt('music/lyrics')
    inctxt('music.*')

=head1 DESCRIPTION

This module processes XML files with an approach similar to OMNIMARK.

Down translation function C<dt> receives a filename and a set of expressions
(functions) defining the processing and associated values for each element.

C<dtstring> is similar but takes input from a string instead of a file.

=head2 C<inctxt> function

C<inctxt(pattern)> is true if the actual element path matches the provided 
pattern. This function is ment to be used in the element functions in order
to achive context dependent processing. 

=head2 User provided element processing functions

The user must provide an HASH with a function for each element,
that computes element output. Functions can use the element name C<$q>, 
the element content C<$c> and the atribute values hash C<%v>. 

All those global variables are defined in C<$main::>.

Each time an element is find the associated function is called.

Content is calculated by concatenation of element contents strings and
interior elements return values.

=head2 C<-default> function

When a element has no associated function, the function associated with 
C<-default> called. If no C<-default> function is defined the default function 
returns a XML like string for the elemente.

=head2 C<-outputenc> option

C<-outputenc> defines the output encoding (default is Unicode UTF8).

=head2 C<-inputenc> option

C<-inputenc> forces a input encoding type. Whenever that is possible,
define the input encoding in the XML file:

   <?xml version='1.0' encoding='ISO-8859-1'?>

=head2 C<-pcdata> function

C<-pcdata> function is used to define tranformation over the contents.
Typically this function should look at context (see C<inctxt> function)

The default C<-pcdata> function is the identity

=head2 C<-begin> function

Function to be executed before processing XML file.

Example of use: inicialization of sife-effect variables

=head2 C<-end> function

Function to be executed after processing XML file.
I can use C<$c> content value.
The value returned by C<-end> will be the C<dt> return value.

Example of use: post-processing of returned contents 

=head2 C<toxml> function

This is the default "-default" function. It can be used to generate xml
based on C<$c> C<$q> and C<%v> variables. Example: add a new attribute to
element C<ele1> without changing it:

  %handler=( ...
             ele1 => sub { $v{at1} = "v1"; toxml(); },
           )

=head1 Skeleton generation

It is possible to build an initial processor program based on an example

To do this use the function C<mkdtskel(filename)>.

Example:

  perl -MXML::DT -e 'mkdtskel "f.xml"' > f.pl

=head1 BUGS

The global variables C<$q %v $c> are defined in main. So you may have to
write C<$::c> ... if you are outside main.

=head1 Author

Jose Joao, jj@di.uminho.pt

  http://www.di.uminho.pt/~jj/perl/XML/

=cut

sub dt {
  my ($file,%xml)=@_;

  # create a new XML::Parser instance using Tree Style
  if ($xml{-inputenc} eq 'ISO-8859-1'){
     $parser = new XML::Parser(Style => 'Tree',
                               ErrorContext => 2 ,
                               ProtocolEncoding => 'ISO-8859-1');
  }
  else { $parser = new XML::Parser(Style => 'Tree',
                                   ErrorContext => 2 ,
                                  );
  }

  #execute Begin action if she exists
  if ($xml{-begin}){ &{$xml{-begin}} }

  # Convert XML to Perl code
  $tree = $parser->parsefile($file);

  if($xml{-end}){ $::c= omni(\%xml,@$tree); 
                  &{$xml{-end}} }
  else          { omni(\%xml,@$tree) }
}

sub ctxt {
  my $level = @_[0];
  $::dtcontext[-$level-1];
}

sub inctxt {
  my $pattern = shift ;
  join("/",@::dtcontext) =~ m!$pattern/[^/]*$! ;
}

sub dtstring
{ my ($string,%xml)=@_;

  # create a new XML::Parser instance using Tree Style
  if ($xml{-inputenc} eq 'ISO-8859-1'){
     $parser = new XML::Parser(Style => 'Tree',
                               ErrorContext => 2 ,
                               ProtocolEncoding => 'ISO-8859-1');
  }
  else { $parser = new XML::Parser(Style => 'Tree',
                                   ErrorContext => 2 ,
                                  );
  }

  #execute Begin action if she exists
  if ($xml{-begin}){ &{$xml{-begin}} }

  # Convert XML to Perl code (Tree)
  $tree = $parser->parse($string);

  if($xml{-end}){ $::c= omni(\%xml,@$tree); 
                  &{$xml{-end}} }
  else          { omni(\%xml,@$tree) }
}

sub omni{
  my ($xml,@l) = @_;
  my $r = "";
  my ($name, $val, @val,$atr, $aux);
  while( @l) {
    ($name, $val, @l) = @l;
    if ($name eq "0"){ 
      $aux= ($xml->{-outputenc} eq 'ISO-8859-1')?lat1::utf8($val): $val ;
      if(defined $xml->{-pcdata}) {
         push(@main::dtcontext,"-pcdata");
         $main::c=$aux; 
         $aux=&{$xml->{-pcdata}}; 
         pop(@main::dtcontext);
      }
      $r .= "$aux";
#         if ($xml->{-outputenc} eq 'ISO-8859-1'){$r.=lat1::utf8($val);}
#         else {$r .= "$val";}
    }
    else  {($atr,@val) = @$val;
           push(@main::dtcontext,$name);
           $r .= omniele($xml, $name, omni($xml,@val), $atr);
           pop(@main::dtcontext);}
  }
  $r;
}

sub omniele {
  my $xml=shift;
  my $aux;
  ($main::q,$main::c,$aux)=@_; 
  %main::v=%$aux;
  if ($xml->{-outputenc} eq 'ISO-8859-1'){
   for (keys %::v){ $::v{$_} = lat1::utf8($::v{$_}) ; }
  }

  if   (defined $xml->{$::q}) {&{$xml->{$::q}} }
  elsif(defined $xml->{'-default'}) {&{$xml->{'-default'} }}
  else {toxml();}
}

sub toxml {"<$::q ".
              join(" ",map {"$_=\"$::v{$_}\""} keys %::v ) .
              ">$::c</$::q>"}

%mkdtskel=( '-default' => sub{$element{$::q}=1; 
                              for (keys %::v){$att{$::q}{$_}=1 }; ""},

             '-end' => sub{ print <<'END';
#!/usr/bin/perl
use XML::DT ;
my $filename = shift;

%handler=(
#    '-outputenc' => 'ISO-8859-1',
#    '-default'   => sub{"<$q>$c</$q>"},
END

for $name (keys %element){
     print "     '$name' => sub{\"\$q:\$c\"},";
     print '# remember $v{',
            join('},$v{',keys %{$att{$name}}),
            '}' if $att{$name}; 
     print "\n";
}
print <<'END';
);
print dt($filename,%handler); 
END
            });

sub mkdtskel{ dt(shift,%mkdtskel)}

package lat1;

=head1 NAME

C<lat1.pm> - module for unicode utf8 to latin1 translation

=head1 SYNOPSIS

   $latin1string = lat1::utf8($utf8string)

=head1 Bugs

Translating the laint1 subset of unicode utf8 is very simples and needs no
tables.

If you need more complex translation, see the perl modules about unicode
and the C<recode> command.

=cut

sub utf8{
  my $t=shift;
  $t =~ s/([])(.)/ $1 eq "" ? chr( ord($2) | 0100): $2 /ge;
  $t;
}

1;
